#include <stdio.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <config.h>
#include <service.h>
#include <socket.h>
#include <netdb.h>

/* macro to swap the bytes in a 32-bit variable */
#define swapbytes32( x ) \
    { \
        uint32_t data = *( uint32_t* )( x ); \
        data = (( data&0xff000000 )>>24 )|    \
               (( data&0x00ff0000 )>>8 )|    \
               (( data&0x0000ff00 )<<8 )|    \
               (( data&0x000000ff )<<24 );     \
        *( uint32_t* )( x ) = data;            \
    }

/* macro to swap the bytes in a 64-bit variable */
#define swapbytes64(x) \
{ \
	uint32_t *words = (uint32_t *)(x); \
	uint32_t temp0;  \
	uint32_t temp1;  \
    temp0 = words[0];    \
    swapbytes32(&temp0);  \
    temp1 = words[1];    \
    swapbytes32(&temp1);  \
    words[1] = temp0;    \
    words[0] = temp1;    \
}


#define SEND_BUFFER_SIZE 1000
static char send_buffer[ SEND_BUFFER_SIZE ];
static int  current_buffer_use;
static int  sock;
static disconnect_cb discCallback = NULL;
static connect_cb connCallback = NULL;
static const char *server_name = NULL;
static int portnumber=0;

#define STORE send_buffer[ current_buffer_use++ ]

static int try_reconnect()
{
    struct addrinfo hints;
    struct addrinfo *result, *rp;
    int getaddr_result;
    int connected;
    char port_str[100];

    if( server_name==NULL || portnumber==0 )
        return 0;

    /* Obtain address(es) matching host/port */

    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = 0;
    hints.ai_protocol = 0;
    snprintf( port_str, 100, "%d", portnumber);
    port_str[99]=0;

    getaddr_result = getaddrinfo( server_name, port_str, &hints, &result);
    if (getaddr_result != 0)
        die(gai_strerror(getaddr_result));

    connected=0;
    for (rp = result; rp != NULL; rp = rp->ai_next)
    {
        sock = socket(rp->ai_family, rp->ai_socktype,
                      rp->ai_protocol);
        if ( sock == -1)
            continue;

        if (connect( sock, rp->ai_addr, rp->ai_addrlen) != -1)
        {
            connected=1;
            break;
        }
    }

    freeaddrinfo( result );

    if( !connected )
        fprintf( stderr, "unable to connect\n");
    else if( connCallback!=NULL ) 
        connCallback();

    return sock;
}

int connectToServer( const char* server,
                     int   port )
{
    sock = -1;
        
    // note that we keep this pointer, the calling app/lib has
    // to make sure it stays where it is
    server_name = server;
    portnumber  = port;
    return try_reconnect();
}

void closeConnection()
{
    close( sock );
    sock = -1;
}

static void set_disconnected()
{
    if( sock == -1 )
            return;

    fprintf( stderr, "disconnected\n");
    closeConnection();
    if( discCallback!=NULL )
        discCallback();
}


double check_and_readSingleRequestResult( const char* ptr, int pos)
{
	double ret=0;
	char *myptr=NULL;
	int error_code, result_type;
	if ( ptr[ pos ] != 8 ) //TYPE_REQUEST_RESULT
	{
		die( "wrong data type, want request_result" );
	}

	myptr = (char*)ptr+pos+3;
	error_code = *myptr;
	myptr++;
	result_type = *myptr;
	myptr++;

	if(error_code != 0) {
		die("error getting request result");
	}

	if(result_type == 1) { //TYPE_SINGLE
		ret = readDouble(myptr, 0);
	}
	else {
		die("wrong request result typ, want type_single");
	}

    return ret;
}


time_line* check_and_readTimeLineRequestResult( const char* ptr, int pos)
{
	char *myptr=NULL;
	time_line* ret=NULL;
	int error_code, result_type;
	int i;

	if ( ptr[ pos ] != 8 ) //TYPE_REQUEST_RESULT
	{
		die( "wrong data type, want request_result" );
	}

	myptr = (char*)ptr+pos+3;
	error_code = *myptr;
	myptr++;
	result_type = *myptr;	//== pos 5
	myptr++;

	if(error_code != 0) {
		die("error getting request result");
	}

	if(result_type == 3) { //TYPE_TIMELINE
		ret = malloc(sizeof(time_line));
		ret->count = readUint64_t(myptr, 0);

		if( ret->count==0 )
		{
			ret->data=NULL;
			return ret;
		}

		ret->data  = malloc( 2*ret->count*sizeof(double));
		if ( ret->data == NULL )
		{
			die( "unable to allocate bytes for array" );
		}

		//ret->data = myptr+sizeof(double);
		myptr+=8;

		for(i=0; i<ret->count; i++)
		{
			((uint64_t*)(ret->data))[2*i] = readUint64_t(myptr, i*16);
			((double*)(ret->data))[2*i+1] = readDouble(myptr, i*16+sizeof(double));
		}
	}
	else {
		printf("result_type: %d\n", result_type);
		die("wrong request result typ, want type_vector");
	}

    return ret;
}


double check_and_readDouble( const char* ptr,
                             int   pos )
{
    if ( ptr[ pos ] != ( char )TYPE_DOUBLE )
    {
        die( "wrong data type, want double" );
    }

    return readDouble( ptr, pos+3 );
}

double readDouble( const char* ptr,
                   int   pos )
{
    union types_union {
        double   d;
        uint64_t u;
    } t;

    t.u =  ( uint64_t )( ptr[ pos+7 ]&0xff )<<56
            |( uint64_t )( ptr[ pos+6 ]&0xff )<<48
            |( uint64_t )( ptr[ pos+5 ]&0xff )<<40
            |( uint64_t )( ptr[ pos+4 ]&0xff )<<32
            |( uint64_t )( ptr[ pos+3 ]&0xff )<<24
            |( uint64_t )( ptr[ pos+2 ]&0xff )<<16
            |( uint64_t )( ptr[ pos+1 ]&0xff )<<8
            |( uint64_t )( ptr[ pos+0 ]&0xff );

    return t.d;
}

uint64_t check_and_readUint64_t( const char* ptr,
                                 int   pos )
{
    if ( ptr[ pos ] != ( char )TYPE_UINT64_T )
    {
        die( "wrong data type, want uint64_t" );
    }

    return readUint64_t( ptr, pos+3 );
}

uint64_t readUint64_t( const char* ptr,
                       int   pos )
{
    return ( uint64_t )( ptr[ pos+7 ]&0xff )<<56
           |( uint64_t )( ptr[ pos+6 ]&0xff )<<48
           |( uint64_t )( ptr[ pos+5 ]&0xff )<<40
           |( uint64_t )( ptr[ pos+4 ]&0xff )<<32
           |( uint64_t )( ptr[ pos+3 ]&0xff )<<24
           |( uint64_t )( ptr[ pos+2 ]&0xff )<<16
           |( uint64_t )( ptr[ pos+1 ]&0xff )<<8
           |( uint64_t )( ptr[ pos+0 ]&0xff );
}

const char* check_and_readPtr( const char *ptr, int pos)
{
    if ( ptr[ pos ] != ( char )TYPE_PTR )
    {
        die( "wrong data type, want ptr" );
    }

    return readPtr( ptr, pos+3 );
}

const char* readPtr( const char *ptr, int pos)
{
    return &ptr[pos];
}

uint32_t check_and_readUint32_t( const char* ptr,
                                 int   pos )
{
    if ( ptr[ pos ] != TYPE_UINT32_T )
    {
        die( "wrong data type, want uint32_t" );
    }

    return readUint32_t( ptr, pos+3 );
}

uint32_t readUint32_t( const char* ptr,
                       int   pos )
{
    return ( ptr[ pos+3 ] )<<24
           |( ptr[ pos+2 ]&0xff )<<16
           |( ptr[ pos+1 ]&0xff )<<8
           |( ptr[ pos+0 ]&0xff );
}

uint16_t readUint16_t( const char* ptr,
                       int   pos )
{
    return ( ptr[ pos+1 ] )<<8
           |( ptr[ pos+0 ]&0xff );
}

void expec_deleteStringList( string_list* to_delete )
{
    int loop;

    if( to_delete->data!=NULL )
    {
        for ( loop = 0; loop < to_delete->count; loop++ )
        {
            free( to_delete->data[ loop ] );
        }
        free( to_delete->data );
    }
    free( to_delete );
}

string_list* check_and_readStringList( const char* ptr,
                                       int   pos )
{
    int          elem_count;
    int          len;
    string_list* list;

    list = malloc( sizeof( string_list ));
    if ( list == NULL )
    {
        die( "unable to allocate memory for new string_list" );
    }
    list->count=0;

    if ( ptr[ pos ] != TYPE_STRINGLIST )
    {
        die( "wrong data type, want string list" );
    }

    pos += 3;
    // 2 byte element count
    elem_count = readUint16_t( ptr, pos );
    pos       += 2;
    list->data = malloc( elem_count*sizeof( char* ));
    if ( list->data == NULL )
    {
        die( "unable to allocate bytes for array" );
    }

    while ( elem_count > 0 )
    {
        // each string has a 2 byte header with the lenght
        len  = readUint16_t( ptr, pos );
        pos += 2;
        // printf("reading element #%d at pos %d of length %d\n", elem_count, pos, len);
        // now get the string, each string has the \0 still
        // on board
        list->data[ list->count++ ] = strdup( &ptr[ pos ] );
        pos                        += len+1;
        elem_count--;
    }
    return list;
}

void expec_deleteTimeLine( time_line* to_delete )
{
    if( to_delete->data!=NULL )
        free( to_delete->data );
    free( to_delete );
}

time_line* check_and_readTimeLine( const char* ptr, int pos )
{
    uint64_t elem_count;
    time_line* timeline;
    uint64_t *conv_old_ptr;
    uint64_t *conv_new_ptr;
    uint64_t num_to_swap;

    if( sizeof(uint64_t)!=sizeof(double) )
    {
        die("sizeof(uint64_t)!=sizeof(double)\n");
    }

    timeline = malloc( sizeof( time_line ));
    if ( timeline == NULL )
    {
        die( "unable to allocate memory for new timeline" );
    }
    timeline->count=0;

    if( ptr[ pos ] != TYPE_TIMELINE )
    {
        die( "wrong data type, want timeline" );
    }

    pos += 3;
    // first 8 bytes a #elements
    elem_count = readUint64_t( ptr, pos );
    timeline->count = elem_count;
    if( elem_count==0 )
    {
        timeline->data=NULL;
        return timeline;
    }

    timeline->data  = malloc( 2*elem_count*sizeof(double));
    if ( timeline->data == NULL )
    {
        die( "unable to allocate bytes for array" );
    }
    pos += sizeof(uint64_t);

    // network layout is 1:1 already
    conv_old_ptr = (uint64_t *) &ptr[pos];
    conv_new_ptr = (uint64_t *) timeline->data;
    for( num_to_swap=0; num_to_swap<elem_count*2; num_to_swap++)
    {
        // we might just swap if needed
        *conv_new_ptr = *conv_old_ptr;
#ifdef ARCH_BIGENDIAN
        swapbytes64(conv_new_ptr);
#endif
        conv_new_ptr++;
        conv_old_ptr++;
    }

    return timeline;
}

void addOutString( const char* ptr )
{
    int len;

    len = strlen( ptr );
    if (( SEND_BUFFER_SIZE-current_buffer_use ) < len+2 )
    {
        die( "send buffer exhausted" );
    }
    strncpy( &send_buffer[ current_buffer_use ], ptr, len+1 );
    current_buffer_use += len+1;
}

void addOutStringList( string_list *v )
{
    int loop;
    char *val;
    int str_len;
        
    // first two bytes are number of elements
    addOutUint16( (uint16_t) v->count );
    for(loop=0; loop<v->count; loop++) {
        val = v->data[loop];
        str_len = strlen(val);
        // header with size
        addOutUint16( (uint16_t) str_len );
        // string itself
        addOutString( val );
    }
}

int string_list_sock_len( string_list *v )
{
	int loop;
	char *val;
	int total_str_len=0;

	// first two bytes are number of elements
    for(loop=0; loop<v->count; loop++) {
        val = v->data[loop];
        // lenght + \0 + 2 byte header (strlen)
        total_str_len += strlen(val)+1+2;
    }

    return total_str_len;
}

void addOutByte( char v )
{
    if (( SEND_BUFFER_SIZE-current_buffer_use ) < 1 )
    {
        die( "send buffer exhausted" );
    }
    STORE = v;
}

void addOutUint16( uint16_t v )
{
    if (( SEND_BUFFER_SIZE-current_buffer_use ) < 2 )
    {
        die( "send buffer exhausted" );
    }
    STORE = ( char )v;
    STORE = ( char )( v>>8 );
}

void addOutUint32( uint32_t v )
{
    if (( SEND_BUFFER_SIZE-current_buffer_use ) < 4 )
    {
        die( "send buffer exhausted" );
    }
    STORE = ( char )v;
    STORE = ( char )( v>>8 );
    STORE = ( char )( v>>16 );
    STORE = ( char )( v>>24 );
}

void addOutUint64( uint64_t v )
{
    if (( SEND_BUFFER_SIZE-current_buffer_use ) < 8 )
    {
        die( "send buffer exhausted" );
    }

    STORE = ( char )v;
    STORE = ( char )( v>>8 );
    STORE = ( char )( v>>16 );
    STORE = ( char )( v>>24 );
    STORE = ( char )( v>>32 );
    STORE = ( char )( v>>40 );
    STORE = ( char )( v>>48 );
    STORE = ( char )( v>>56 );
}

void addOutDouble( double v )
{
    union { 
            double   d;
            uint64_t u;
    } field;
    field.d = v;
    addOutUint64( field.u );
}

void clearOutputBuffer()
{
    current_buffer_use = 0;
}

void socketSendAll()
{
    if( sock == -1 )
        try_reconnect();
    if( sock == -1 )
        return;
    if ( send( sock, send_buffer, current_buffer_use, 0 ) != current_buffer_use )
    {
        fprintf( stderr, "send() sent a different number of bytes than expected\n" );
        set_disconnected();
    }
}

int readBytesTimeout( int   num_to_read,
                      char* ptr,
                      int   sleep_usec )
{
    int num_read = 0;
    int read_chunk;
    fd_set rfds;
    struct timeval tv;
    int retval;

    if( sock == -1 )
        try_reconnect();
    if( sock == -1 )
        return 0;

    FD_ZERO(&rfds);
    FD_SET( sock, &rfds);
    tv.tv_sec  = 0;
    tv.tv_usec = sleep_usec;

    retval = select( sock+1, &rfds, NULL, NULL, &tv);

    if (retval == -1)
    {
        die("select() for network socket failed");
    }
    else if (retval==0)
    {
        return 0;
    }

    while ( num_read < num_to_read )
    {
        read_chunk = recv( sock, &ptr[num_read], num_to_read-num_read, MSG_WAITALL );
        if ( read_chunk <= 0 )
        {
            set_disconnected();
            return 0;
        }
        num_read += read_chunk;
    }

    return num_to_read;
}

void readBytes( int   num_to_read,
                char* ptr )
{
    int num_read = 0;
    int read_chunk;

    if( sock == -1 )
        try_reconnect();
    if( sock == -1 )
        return;
 
    while ( num_read < num_to_read )
    {
        read_chunk = recv( sock, &ptr[num_read], num_to_read-num_read, MSG_WAITALL );
        if ( read_chunk <= 0 )
        {
            set_disconnected();
            return;
        }
        num_read += read_chunk;
    }
}

char* readMessage( uint32_t* size, int timeout_usec )
{
    char* size_ptr = ( char* )size;
    char* message_ptr;

    if( readBytesTimeout( 4, size_ptr, timeout_usec) == 0 )
    {
        return NULL;
    }
#ifdef ARCH_BIGENDIAN
    swapbytes32( size_ptr );
#endif
    message_ptr = ( char* )malloc( *size );
    if ( message_ptr == NULL )
    {
        fprintf( stderr, "malloc for %u bytes\n", *size );
        exit( 1 );
    }
    /* printf("incomming with size %d\n", *size); */
    readBytes(( *size )-4, message_ptr );
    return message_ptr;
}

void set_disconnect_notifier( disconnect_cb cb )
{
    discCallback = cb;
}

void set_connect_notifier( connect_cb cb )
{
    connCallback = cb;
}
