#ifndef _HAVE_SOCKET_H_
#define _HAVE_SOCKET_H_

#include <dataheap_misc.h>
#include <stdint.h>

/**
 * type ids to know what to unpack
 */
typedef enum {
    TYPE_CHAR       = 1,
    TYPE_UINT32_T   = 2,
    TYPE_UINT64_T   = 3,
    TYPE_PTR        = 4,
    TYPE_DOUBLE     = 5,
    TYPE_STRINGLIST = 6,
    TYPE_TIMELINE   = 7
} RPC_PARAMETER_TYPES;

/** a type to pack the timeline data
 *  data has 2*count entries with
 *  {time1,value1,time2,value2,...}
 */
typedef struct time_line_t {
    char *data;
    uint64_t count;
} time_line;

#define MESSAGE_ENVELOPE_SIZE     9
#define PARAMETER_ENVELOPE_SIZE   3

typedef void (*disconnect_cb)( void );
typedef void (*connect_cb)( void );

int connectToServer( const char *server, int port);
void closeConnection();
double check_and_readDouble( const char *ptr, int pos);
double readDouble( const char *ptr, int pos);
uint64_t check_and_readUint64_t( const char *ptr, int pos);
double check_and_readSingleRequestResult( const char* ptr, int pos);
time_line* check_and_readTimeLineRequestResult( const char* ptr, int pos);
uint64_t readUint64_t( const char *ptr, int pos);
uint32_t check_and_readUint32_t( const char *ptr, int pos);
uint32_t readUint32_t( const char *ptr, int pos);
uint16_t check_and_readUint16_t( const char *ptr, int pos);
uint16_t readUint16_t( const char *ptr, int pos);
const char* check_and_readPtr( const char *ptr, int pos);
const char* readPtr( const char *ptr, int pos);
void expec_deleteStringList( string_list *to_delete );
void expec_deleteTimeLine( time_line *to_delete );
string_list *check_and_readStringList( const char *ptr, int pos);
time_line *check_and_readTimeLine( const char *ptr, int pos);
void addOutString( const char *ptr );
void addOutByte( char v );
void addOutUint16( uint16_t v );
void addOutUint32( uint32_t v );
void addOutUint64( uint64_t v );
void addOutDouble( double v );
void addOutStringList( string_list *v );
int string_list_sock_len( string_list *v );
void clearOutputBuffer();
void socketSendAll();
void readBytes( int num_to_read, char *ptr);
char *readMessage( uint32_t *size, int timeout_usec);
void set_disconnect_notifier( disconnect_cb cb );
void set_connect_notifier( connect_cb cb );

#endif /* _HAVE_SOCKET_H_ */
