/**
 * eeMark
 * http://eeclust.de
 *
 * Copyright (c) 2011-2012, ZIH, TU Dresden, Federal Republic of Germany
 *
 * See the file COPYING in the package base directory for details
 **/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "energymeter_interface.h"

#define SEPARATOR ";"
#define MAX_CONNECTIONS 20

typedef struct {
	char * host;
	int port;
} server_t;

char mark[256];
server_t * servers = NULL;
int srv_count = 0;
int verbose = 0;
int sockfd[MAX_CONNECTIONS];

void exit_error(char * errmsg) {
	fprintf(stderr, "SpecPowerLib ERROR: %s\n", errmsg);
	exit(EXIT_FAILURE);
}

static int sendStr(const char * buffer, int sockno) {
	int r;
	if(sockno < 0 || sockno > MAX_CONNECTIONS-1 || buffer==NULL || sockfd[sockno] < 0) {
		return -1;
	}

	r = write(sockfd[sockno], buffer, strlen(buffer));
	if(r<0) {
		//printf("could not write data to socket");
		;
	}
	return r;
}

/*
static void get_meter_responses() {
	char buffer[256];
	int i;
	bzero(buffer,256);
	for(i=0; i<MAX_CONNECTIONS; i++) {
		if (sockfd[i] >= 0) {
			while(read(sockfd, buffer, 255) != 0) {
				printf("Meter says: %s\n", buffer);
				bzero(buffer,256);
			}
		}
	}
}
*/

void em_start_measurement() {
	char buffer[256];
	int i;

	for(i=0; i<MAX_CONNECTIONS; i++) {
		if (sockfd[i] >= 0) {
			sprintf(buffer, "Go,0,0,%s\r\n", mark);
			sendStr(buffer, i);
			bzero(buffer,256);
			read(sockfd, buffer, 255);
			//printf("Start: %s\n", buffer);
		}
	}
}

void em_stop_measurement() {
	char buffer[256];
	int i;

	for(i=0; i<MAX_CONNECTIONS; i++) {
		if (sockfd[i] >= 0) {
			sendStr("Stop\r\n", i);
			bzero(buffer,256);
			read(sockfd, buffer, 255);
			//printf("Stop: %s\n", buffer);
		}
	}
}


int em_connect() {
        struct sockaddr_in saddr;
        struct hostent *server;
	char buffer[256];
	int i, err_count=0;

	for(i=0; i<srv_count; i++) {

		server = gethostbyname(servers[i].host);
		if (server==NULL) {
		        return 0; //failure
		}

		bzero((char*) &saddr, sizeof(saddr));
		saddr.sin_family = AF_INET;
		bcopy((char*)server->h_addr, (char*)&saddr.sin_addr.s_addr, server->h_length);
		saddr.sin_port = htons(servers[i].port);

		printf("connecting to %s:%d...\n", servers[i].host, servers[i].port);

		if ((sockfd[i] = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		        //exit_error("could not open socket");
		        err_count++; //failure
			continue;
		}
		if(connect(sockfd[i], (struct sockaddr *)&saddr, sizeof(saddr)) == -1) {
		        //printf("error connecting");
			close(sockfd);
		        err_count++; //failure
			continue;
		}

		//ask PTD about the meter that's attached
		if(sendStr("Hello\r\n", i)) {
			read(sockfd[i], buffer, 255);
			if(strncmp(buffer, "Hello, PTDaemon here!", 21) != 0) {
			        err_count++; //failure
				continue;
			}
		}
	}

	if(err_count == 0) {
		return EM_CONNECTED; //success
	}
	else {
		if(verbose>0) {
			printf("SpecPowerLib VERBOSE: %d errors while trying to connect to ptd daemons\n", err_count);
		}
		return EM_NOT_CONNECTED;
	}
}

void em_init() {
	char *addr=NULL, *ptr=NULL;
	char tmp[256];
	int i;
	
	for(i=0; i<MAX_CONNECTIONS; i++) {
		sockfd[i] = -1;
	}

	//get address from ENV ("host:port")
	addr = getenv("SPLIB_SERVER");
	if(addr==NULL) {
		exit_error("environment variable \"SPLIB_SERVER\" not set.");
	}

	//first, split the list of addresses
	ptr = strtok(addr, SEPARATOR);
	while(ptr != NULL) {
		int len = strlen(ptr);
		servers = realloc(servers, ++srv_count * sizeof(server_t)); //increase counters size
		servers[srv_count-1].host = malloc(len * sizeof(char));
		bzero(servers[srv_count-1].host, len * sizeof(char));
		strncpy(servers[srv_count-1].host, ptr, len); //store complete addr in server field first
		ptr = strtok(NULL, SEPARATOR);
	}

	//then split host:port
	for(i=0; i<srv_count; i++) {
		ptr=strtok(servers[i].host, ":");
		if(ptr!=NULL) {
			ptr = strtok(NULL, ":");
			if(ptr==NULL) {
				exit_error("no port given in one of the entries in SPLIB_SERVER. Use format: \"host:port\".");
			}
			servers[i].port = atoi(ptr);
			*(ptr-1) = '\0';

			//printf("DEBUG... i=%d host=%s port=%d\n", i, servers[i].host, servers[i].port);
		}
		else {
			exit_error("invalid address in \"SPLIB_SERVER\".");
		}
	}

	//get counter from ENV
	strcpy(mark, "eeMark");
	ptr = getenv("SPLIB_MARK");
	if(ptr!=NULL) {
		strcpy(mark, ptr);
	}


	//optional verbose setting
	ptr = getenv("SPLIB_VERBOSE");
	if(ptr!=NULL) {
		verbose = atoi(ptr);
	}
}

void em_close() {
	int i;
	for(i=0; i<MAX_CONNECTIONS; i++)
		if(sockfd >= 0)
			close(sockfd);
}


double em_get_energy(uint64_t start, uint64_t stop) {
	char buffer[256];
	double t = 0.0, dummy, sum = 0.0;
	uint64_t count;
	int i;

	for(i=0; i<MAX_CONNECTIONS; i++) {
		if(sockfd[i] < 0) continue;

		bzero(buffer,256);
		sendStr("Watts\r\n", i);

		read(sockfd[i], buffer, 255);
		sscanf(buffer, "Watts,%lf,%lf,%lf,%llu", &t, &dummy, &dummy, (unsigned long long*)&count);

		if(verbose>0) {
			printf("SpecPowerLib VERBOSE: Socket #%d, %llu element(s), duration = %.3lf ms, average = %.3lf W\n", i, (unsigned long long)count, (double)(stop-start)/1000, t);
		}

		sum += (double)(stop-start) * t / 1000;
	}

	return sum;
}

double em_get_power(uint64_t start, uint64_t stop) {
	char buffer[256];
	double t = 0.0, sum = 0.0;
	int i=0;

	for(i=0; i<MAX_CONNECTIONS; i++) {
		if(sockfd[i] < 0) continue;

		bzero(buffer,256);
		sendStr("Watts\r\n", i);

		read(sockfd[i], buffer, 255);
		sscanf(buffer, "Watts,%lf", &t);
		sum += t;
	}

	return sum;
}

